/* ComponentInfo.java - hades.manager.ComponentInfo
 *
 * read, write, manage, GUI to edit Colibri CLB files
 *
 * 10.03.01 - allow multiline [description] data
 * 09.03.01 - integration into Hades 0.87, FNH
 * 01.07.98 - original version by
 * @author    Ulrich Dallmann       (2dallman@informatik.uni-hamburg.de)
 * @author    Karola Kroenert       (2kroene@informatik.uni-hamburg.de
 */

package hades.manager;

import java.awt.Image;
import java.io.*;
import java.util.*;
import javax.swing.ImageIcon;

/**
 * ComponentInfo - A class to read, write, and manager Colibri CLB data.
 *
 * The ComponentInfo data is used to store dependency information,
 * author, version, name attributes, and a short description text
 * about Hades simulation models and design files.
 *
 * <p>
 * The information is is stored as plain text in a '.clb'-file
 * which stands for CoLiBri or Component_Library_Browser.
 * For loading existing and creating new information files,
 * two different constructors are provided.
 * <p>
 * The format of the file is as follows:
 * every information consists of a tag line and the data line(s),
 * tags are enclosed in [], 
 * some tags are a MUST while others are optional.
 * The currently supported tags are:
   <pre>
   [files x]       MUST  x lines with all filenames used by the component
   [references x]  MUST  x lines contain the URLs of all used sub-components
   [type]          MUST    the following line is the type of the component
   [archive]       OPT.    the following line is the filename of a jar-archive containing ALL the files listed here
   [start]         OPT.    name of the start class (sunw.demo.juggler.juggler)
   [init]          OPT.    the parameter of an init-method (eg. a filename)
   [icon]          OPT.    the filename of the component's icon
   [image]         OPT.    the component's (larger) image
   [html]          OPT.    the filename of a html file 
   [description]   OPT.    the following line (only one!) is a component description with encoded control chars
    </pre>
 * @version    1/7/1998
 * @author    Ulrich Dallmann       (2dallman@informatik.uni-hamburg.de)
 * @author    Karola Kroenert       (2kroene@informatik.uni-hamburg.de
 */

public class ComponentInfo 
{
 
  public static boolean debug = true;
 
  private String name       = "unnamed";
  private String author     = "--";
  private String version    = "1.001";
  private String type       = null;
  private String html       = null;
  private String description= null;

  private String pathname   = "no pathname";
  private String iconpath   = "";
  private String imagepath  = "";
  private String start      = null;
  private String archive    = null;
  private String init       = null;

  private Vector files = new Vector();
  private Vector refs  = new Vector();

  
  
  /**
   * Creates an empty ComponentInfo for new Components
   * To store information call the set..() methods
   * and finally saveComponentInfo()
   */
  public ComponentInfo()
  {
  }
  
  /**
  * Creates a new ComponentInfo and loads all the supported
  * data out of the given file.
  * 
  * @param      server      server
  * @param      basedir      base directory
  * @param      relpath      relative path INCLUDING the name
  * @exception  IOException    if an Exception occurs while loading the file
  * @see                java.io.File
  */
  public ComponentInfo( String pathname ) 
    throws FileNotFoundException, IOException
  {
    if (debug) msg( "-I- ComponentInfo.<init>: " + pathname );
    this.pathname = pathname;

    // ask the DesignManager to read the pathname data into a buffer
    //
    ByteArrayOutputStream baos = 
                     DesignManager.
                       getDesignManager().
                         readStreamIntoBuffer( pathname );

    // msg( "-#- BAOS = " + baos + "\n" );

    // extract information
    BufferedReader br = new BufferedReader(new StringReader( baos.toString() ));
    String line;
    while ((line=br.readLine())!=null)
    {
      line=line.trim().toLowerCase();
      if (line.startsWith("[files "))
      {
        int num=0;
        try
        { 
          num = Integer.parseInt(line.substring(7,line.indexOf(']'))); 
        }  
        catch (Exception ex)
        {
          msg("-E- ComponentInfo: invalid number of [files x]: "+ line );
          throw new IOException("invalid 'clb'-file");
        }
        for (int i=0; i<num; i++)
        {
          files.addElement(br.readLine().trim());
        }
      }
      else if (line.startsWith("[references "))
      {
        int num=0;
        try
        { 
          num = Integer.parseInt(line.substring(12,line.indexOf(']'))); 
        }  
        catch (Exception ex)
        {
          msg("-E- ComponentInfo: invalid number of [references x]: "+ line );
          throw new IOException("invalid 'clb'-file");
        }
        for (int i=0; i<num; i++)
        {
          refs.addElement(br.readLine().trim());
        }
      }
      else if (line.equals("[type]"))
      {
        type = br.readLine().trim();
      }
      else if (line.equals("[start]"))
      {
        start = br.readLine().trim();
      }
      else if (line.equals("[init]"))
      {
        init = br.readLine().trim();
      }
      else if (line.equals("[archive]"))
      {
        archive = br.readLine().trim();
      }
      else if (line.equals("[icon]"))
      {
        iconpath  = br.readLine().trim();
      }
      else if (line.equals("[image]"))
      {
        imagepath = br.readLine().trim();
      }
      else if (line.equals("[author]"))
      {
        author = br.readLine().trim();
      }
      else if (line.equals("[version]"))
      {
        version = br.readLine().trim();
      }
      else if (line.equals("[description]"))
      {
        StringBuffer sb = new StringBuffer();
        while ((line=br.readLine())!=null) {
          if (line.equals( "[end description]" )) continue; // next outer iter.
          else {
            sb.append( line );
            sb.append( "\n" );
          }
        }
        description = sb.toString();
      }
      else if (line.equals("[html]"))
      {
        html = br.readLine().trim();
      }
    }
    if (files.size()==0) msg("-W- ComponentInfo -> Component without any files!");
    // loadIcon();
    // loadImage();
  }


  // deprecated !!!
  public String getSubtype()
  {
    String ret = null;
    if (!type.equals("beans")) return ret;
    if (init == null) ret = "simobject";
    else if (init.endsWith(".hds")) ret = "design";
    return ret;
  }
  
  
  public String getName()
  {
    return name;
  }
 
 
  public void setName(String name)
  {
    this.name = name;
  }


  /**
   * return the component's "pathname" URL. For example:
   * zip://hades.zip?/libraries/gatelib/And2.clb
   * file://c:/windows/temp/hades/praktikum/dlatch.clb
   */
  public String getPathname() {
    return pathname;
  }

  public void setPathname( String pathname ) {
    this.pathname = pathname;
  }


  /**
   * Get filenames that belong to this component.
   *
   * @return      Vector  Vector containing the filenames
   */
  public Vector getFiles()
  {
    return files;
  }

  /**
   * Add a new file that belongs to this component.
   *
   * @param    file    filename as String
   */
  public void addFile(String file)
  {
    if (file!=null) files.addElement(file);
  }
 
 
  /**
   * Remove a component's file.
   *
   * @param   file    filename
   * @return       boolean  whether file was found and removed
   */
  public boolean removeFile(String file)
  {
    return files.removeElement(file);
  }



  /**
   * Get Reference-URLs of sub-components used by this component.
   *
   * @return      Vector  Vector containing the URLs
   */
  public Vector getReferences()
  {
    return refs;
  }
  
  /**
   * Add a new Reference-URL.
   *
   * @param    ref   Reference-URL
   */
  public void addReference(String ref)
  {
    if (ref!=null) refs.addElement(ref);
  }
  
  /**
   * Remove a Reference-URL.
   *
   * @param   ref       Reference-URL
   * @return       boolean  whether url was found and removed
   */
  public boolean removeReference(String ref)
  {
    return refs.removeElement(ref);
  }
  
  
  /**
   * Get the start class/file for the component.
   *
   * @return      String  the name of the start class/file or null if unknown
   */
  public String getStart()
  {
    return start;
  }
  
  /**
   * Set the start/main class for the component.
   *
   *    @param  startClass     name of the start class/file as String
   */
  public void setStart(String start)
  {
    this.start = start;
  }
  
  
  /**
   * Get the parameter for a possible init method (eg. a resource-filename)
   *
   * @return      String  the parameter string or null if unknown
   */
  public String getInit()
  {
    return init;
  }
  
  /**
   * Set the parameter for a possible init method (eg. a resource-filename)
   *
   *    @param  init     the parameter string
   */
  public void setInit(String init)
  {
    this.init = init;
  }


  /**
  * Get the type of the component.
  *
  * @return      String  a component-type (as defined in ComponentType or custom)
  * @see        ComponentType
  */
  public String getType()
  {
    return type;
  }
  
  
  
  /**
  * Set the type of the component.
  *
  * @param    type    a component-type (as defined in ComponentType or custom)
  * @see        ComponentType
  */
  public void setType(String type)
  {
    this.type = type;
  }
  
  
  /**
  *
  */
  public String getAuthor()
  {
    if (author != null) return author;
    else return "unknown";
  }
  
  
  /**
  * Set the component's author.
  *
  *    @param  author     author of the component
  */
  public void setAuthor(String author)
  {
    this.author = author;
  }



  public String getVersion() {
    return version;
  }

  public void setVersion( String s ) {
    version = s;
  }


  /**
  * Get the archive name for the component. All
  * other files are searched in this jar-archive.
  *
  * @return      String  archive name
  */
  public String getArchive()
  {
    return archive;
  }
  
  /**
  * Set the archive name for the component. All
  * other files are searched in this jar-archive.
  *
  * @param    archive       archive name
  */
  public void setArchive(String archive)
  {
    this.archive = archive;
  }
  
  
  /**
  * Get the pathname for the thumbnail image for the component. 
  *
  * @return String  filename of an image
  */
  public String getImagepath()
  {
    return imagepath;
  }
  
  /**
  * Set the pathname for the thumbnail image for the component. 
  *
  * @param  image   filename of an image
  */
  public void setImagepath(String imagepath)
  {
    this.imagepath = imagepath;
  }
  
  
  /**
  * Get the iconpath for the component. The iconpath
  * should be a small logo of the component.
  *
  *    @return String  filename of an image
  */
  public String getIconpath()
  {
    return iconpath;
  }
  
  /**
  * Set the iconpath for the component. The iconpath
  * should be a small logo of the component.
  *
  *    @param  iconpath    filename of an image
  */
  public void setIconpath(String iconpath)
  {
    this.iconpath = iconpath;
  }
  
  
  /**
  * Get the description for the component. The
  *    description may contain only ASCII chars
  * (8 bit) but can have multiple lines.
  *
  * @return      String  description as String
  */
  public String getDescription()
  {
    return description;
  }
  
  /**
  * Set the description for the component. The
  *    description may contain only ASCII chars
  * (8 bit) but can have multiple lines.
  *
  * @param    description     text-description
  */
  public void setDescription(String description)
  {
    this.description = description;
  }

  /**
  * Get filenames that contain html information about this component.
  *
  * @return      Vector  Vector containing the filenames
  */
  public String getHTML()
  {
    return html;
  }

  /**
  * Set the file that contains information about this component.
  *
  * @param    file    filename as String
  */
  public void setHTML(String file)
  {
    html = file;
  }
  
  // to be continued for every new tag ...
  
  /** 
  * will return true when "Open" in context menue should be enabled
  * @return (boolean) enable open in context menue
  */
  public boolean showOpen()
  {
    return ((init!=null) && (init.endsWith(".hds")));
  }
  
  
  public boolean search(Vector keys)
  {
    for (int i=0;i<keys.size();i++)
    {
      String key = ((String)keys.elementAt(i)).toLowerCase();
      if (name.toLowerCase().indexOf(key)>=0) return true;
      if (author.toLowerCase().indexOf(key)>=0) return true;
      if (description != null)
        if (description.toLowerCase().indexOf(key)>=0) 
          return true;
    }
    return false;
  }
 
 
  /**
   * Get the type-icon for the component. Delegate to DesignManager.
   *
   * @return    ImageIcon  Icon of this component
   */
  public ImageIcon getIconObject()
  {
    ImageIcon icon = null;
    Image tmp = DesignManager.getDesignManager().getThumbnail( iconpath );
    if (tmp != null) return new ImageIcon( tmp );
    else             return null;
  }
  
  
  /**
   * Get the thumbnail image for the component. 
   * We just delegate this to DesignManager.
   *
   * @return    Image  Image of this component
   */
  public Image getImageObject()
  {
    return DesignManager.getDesignManager().getThumbnail( imagepath );
  }



  /**
   * Save the ComponentInfo information. This uses the original pathname
   * for this CLB file (if any).
   *
   * @exception IOException if an Exception occurs while saving the data
   */
  public void saveComponentInfo()
    throws IOException
  {
    saveComponentInfo( null );
  }
  
  /**
  * Save the ComponentInfo information to the location specified.
  * To follow the standard, the pathname should end with ".clb"
  *
  * @param  pathname   the pathname to store the information in
  * @exception IOException    if an Exception occurs while saving the data
  */
  public void saveComponentInfo( String pathname ) 
    throws IOException
  {
    String  newPathname;
    if (pathname != null) newPathname = pathname;
    else                  newPathname = this.pathname;

    OutputStream OS = null;
    OS = DesignManager.getDesignManager().
           getOutputStream( this, newPathname );

    // assemble the whole file in a string first
    String data = "";
    
    // create [files x] tag
    data += "[files ";
    data += files.size() + "]\n";
    for (int i=0; i<files.size(); i++)
    {
      data += (String)files.elementAt(i) + "\n";
    }
    data += "\n";

    // create [versions x] tag
    data += "[versions 1]\n";
    data += version + "\n\n";

    // create [references x] tag
    data += "[references ";
    data += refs.size() + "]\n";
    for (int i=0; i<refs.size(); i++)
    {
      data += (String)refs.elementAt(i) + "\n";
    }
    data += "\n";
    
    // create [type] tag
    data += "[type]\n" + type + "\n\n";
    
    // create [start] tag
    if (start != null)
    {
      data += "[start]\n";
      data += start + "\n\n";
    }
    
    // create [init] tag
    if (init != null)
    {
      data += "[init]\n";
      data += init + "\n\n";
    }
    
    // create [icon] tag
    if (iconpath != null)
    {
      data += "[icon]\n";
      data += iconpath + "\n\n";
    }
    
    // create [image] tag
    if (imagepath != null)
    {
      data += "[image]\n";
      data += imagepath + "\n\n";
    }
    
    // create [archive] tag
    if (archive != null)
    {
      data += "[archive]\n";
      data += archive + "\n\n";
    }
    
    // create [author] tag
    if (author != null)
    {
      data += "[author]\n";
      data += author + "\n\n";
    }

     //create [description] tag
     if (description != null)
     {
       data += "[description]\n";
       data += description + "\n[end description]\n";
     }
    
    // create [html] tag
    if (html != null)
    {
      data += "[html]\n";
      data += html + "\n\n";
    }

    
    // try to open the file and store the data in it
    try
    {
      DataOutputStream dos = new DataOutputStream( OS );
      dos.write(data.getBytes());
      dos.close();
    }
    catch (IOException ex)
    {
      msg("-E- ComponentInfo: saveComponent failed for: "+ pathname );
      throw new IOException("Failed to save 'clb': " + pathname );
    }
  }
  
  
  
  /**
   * Returns the name of the component which is shown in the JTree!
   */
  public String toString()
  {
    return "CLB:" + pathname + " " + imagepath + " " + iconpath;
  }
  
  /** 
   * send a msg to the Console 
   */
  public static void msg( String s ) { 
    //hades.manager.Console.getConsole().msg( s );
    System.out.println( s );
  }


  public static void main( String argv[] ) throws Exception {
    msg( "-I- ComponentInfo selftest..." );

    ComponentInfo clb[] = new ComponentInfo[ argv.length ];
    for( int i=0; i < argv.length; i++ ) {
      msg( "" );
      msg( "...checking: " + argv[i] );
      clb[i] = new ComponentInfo( argv[i] );
      msg( "...loaded: " + clb[i].toString() );
      msg( "...description: \n" + clb[i].getDescription() );
    }

  }
}


