/* DesignManager.java -- hades.manager.DesignManager
 *
 * Preliminary implementation of the central HADES DesignManager class,
 * responsible for loading, saving, and browsing of simulation models,
 * Designs, and their symbols.
 *
 * 22.01.02 - more attempts to avoid JDK crashes on Windows
 * 21.01.02 - don't re-use FileDialog to avoid JDK1.3.1/Windows crashes
 * 21.01.02 - some progress with lookForInputStream, classpath-resource-hack
 * 14.01.02 - added checkAndAddFilenameExtension utility method
 * 21.12.01 - support both Netscape/MSIE and "traditional FileDialog" 
 * 09.11.01 - implemented getInputStreamFromRelativeName
 * 09.11.01 - use JDK 1.1 methods in getFilename and getDirectory
 * 09.11.01 - check for null filename in selectFileOrURLName
 * 09.11.01 - use SetupManager to initialize the debug flag
 * 18.09.01 - implemented new version of getInputStream, tokenizing
 * 14.09.01 - removed old FileDialog and designNameTable stuff
 * 11.09.01 - next attempt: allow absolute names, don't truncate automatically
 * >>>>>>>> 
 * 16.03.01 - getDesignXXX
 * 12.03.01 - started the full redesign...
 * 05.03.01 - first ZipFile methods
 * 03.03.01 - getThumbnail
 * >>>>>>>>
 * 02.08.99 - workaround Netscape/MSIE bugs for applet use
 * 08.06.99 - encodeWithUnicodEscapes
 * 03.06.99 - added getResourceAsStream method
 * 28.05.99 - use Design.hashCode() for the name/dir tables, more documentation
 * 27.05.99 - set FileDialog.LOAD/SAVE mode, new FileDialog titles
 * 15.04.98 - changed all FileOutputStreams to OutputStreams (->Colibri)
 * 23.03.98 - added getClassURL() method
 * 30.12.97 - getNewDesign() as Wrapper to the Design constructor
 * 27.11.97 - getFileOutputStreamForSymbol to select symbol file names
 * 11.11.97 - 'designNameTable' for better 'save file'
 * 30.09.97 - handle 'Cancel' from the FileDialog
 * 29.09.97 - 'file save' without confirmation, but with auto-backup
 * 26.09.97 - one persistent file dialog for load & save
 * 02.08.97 - first try
 *
 * (C) F.N.Hendrich, hendrich@informatik.uni-hamburg.de
 */

package hades.manager;

import hades.simulator.SimObject;
import hades.models.Design;
import hades.models.FigObject;
import hades.gui.Editor;
import hades.gui.SelectFileOrURLDialog;
import hades.utils.NameMangler;
import hades.utils.EndsWithFilter;
import jfig.utils.SetupManager;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.net.*;

/**
 * DesignManager - the central HADES design manager class.
 * The singleton instance of this class is
 * responsible for loading, saving, and browsing of simulation models,
 * Designs, and their symbols.
 * DesignManager accesses  
 * designs and design resources from local files, JAR/Zip-files, 
 * URLs, and the Java classpath.
 * <p>
 * This class currently supports the following use-cases and scenarios:
 <ul>
  <li>Selecting a file or URL name, either for reading or writing ("SELECT"):
      This displays a dialog window and prompts the user to select 
      a resource file.
      It allows to preset a file/URL name, uses a FilenameFilter 
      when appropriate, and takes the default mode (load or save).
      When closed, the dialog returns a String with the selected file name,
      or null when canceled.
  <li>Saving a new design ("SAVE AS"):
      Prompt the user for a file/URL name for the current Hades design 
      (".hds"), then get an OutputStream to write the data.
  <li>Saving design or SimObject resource data ("SAVE AS"):
      Prompt the user for a file/URL name for the data in question 
      (e.g. ROM initialization data, with a ".rom" extension), 
      then get an OutputStream to write the data. 
      This method will use the file/URL name in its raw form, 
      e.g. an absolute path name. No automatic conversion to relative 
      or classpath-resource names
      is done here.

  <li>Creating a symbol for the current design ("SAVE SYMBOL"):

  <li>Selecting and loading an existing design ("SELECT AND LOAD"):
      Prompt the user to select an existing design file/URL name, 
      then get an InputStream for the design and load it.

  <li>Loading an existing design file ("LOAD"):
      Given an absolute or relative file name, an URL, 
      or a ZIP/JAR/CLASSPATH resource name, get an InputStream 
      for the design and load it.
      In hierarchical designs, also use the information about 
      the parent/toplevel design file/URL name to locate subdesigns. 
      This function tries to be as error-tolerant as possible, even trying
      to recurse into local directories in order to find design files.

  <li>Loading a SimObject or design file resource ("LOAD"):
      Load design resource files, e.g. RAM/ROM initialization data, 
      stimuli data, image data, SimObject symbol files, FIG files 
      in symbols, microcontroller configuration data, etc.
      This method locates resources from absolute or relative file names, 
      URLs, ZIP/JAR/CLASSPATH resources, and tries to be as error-tolerant 
      as possible.
      When necessary, this method tries to complete relative file names via
      trying combinations with the parent design or toplevel design names. 

  <li>Loading an Image or image thumbnail:
      The same as loading a resource, but relying on jfig ImageHelper or
      AWT Toolkit to create an Image object from the raw image data.

  <li>Loading SimObjects ("SIMOBJECT"):
      Load a SimObject from either a given a Java class name 
      (hades.models.gates.And3), or a file/URL name. 
      The latter requires the use of an user-defined ClassLoader 
      and is not implemented yet (import the Colibri stuff here).

  <li>Versioning when writing data ("CHECK-EXISTS-RENAME"):
      Given a file/URL/resource name, check whether the requested location
      already exists. If so, automatically rename the existing data, 
      before writing the new data.

  <li>Check when overwriting existing files ("CHECK-EXISTS-OVERWRITE"):
      Given a file/URL/resource name, check whether the requested location
      is exists. If so, display a dialog to prompt the user to overwrite
      the exisiting data or to cancel the operation.

  <li>ConvertToRelativeOrResourceNames ("CONVERT"):
      Given an existing top-level design file, load that design and convert
      all absolute file names used in subdesigns or SimObject resources to
      relative names (relative to the top-level design).
      If used with an appropriate directory structure for the subdesigns
      and (sub-)design resources, this allows to move a complete hierarchical
      design to other directories or other machines.
      Alternatively, convert all file names to Java classpath resource names,
      allowing to load a design from the Java classpath.     
 </ul>
 */
public class DesignManager  {

  private  static DesignManager  _designManager = null;
  private  static boolean   debug = false;


  private hades.gui.SelectFileOrURLDialog  selectDialog; // Netscape/MSIE style
  private java.awt.FileDialog              fileDialog;   // traditional style
  private java.awt.Frame                   frame; // parent for the dialogs
  private String                           _lastFileDialogDirectory;

  private Hashtable _zipfileTable = new Hashtable();
  private Hashtable thumbnailTable = new Hashtable();




  /**
   *  Return a reference to the Singleton DesignManager.
   */
  public synchronized static DesignManager  getDesignManager() {
    if (_designManager == null) {
       _designManager = new DesignManager();
    }
    return _designManager;
  }


  /**
   *  As DesignManager is a Singleton class, the constructor is private.
   *  Use 'getDesignManager()' to reference the global DesignManager.
   */
  private DesignManager() {
    if (debug) msg( "-I- DesignManager() constructed..." );
    debug = SetupManager.getBoolean( "Hades.DesignManager.Debug", false );
  }



  /**************************************************************/ 
  /*   basic SimObject / Design accessor methods                */
  /**************************************************************/ 


  /**
   * getSimObject(): given a Java classname like "hades.models.gates.And2",
   * return an instance of that class.
   * @exception SimObjectNotFoundException if the requested Object is not found
   */
  public SimObject getSimObject( String classname ) 
         throws SimObjectNotFoundException {
    if (debug) msg( "-I- DM.getSimObject(" + classname + ")" );

    SimObject tmp = null;
    try {
      Class cl = Class.forName( classname );
      tmp = (SimObject) cl.newInstance();
    }
    catch( Exception e ) {
      msg( "-E- DesignManager: SimObject not found: " + classname );
      jfig.utils.ExceptionTracer.trace( e );
      throw new SimObjectNotFoundException( classname );
    }
    return tmp;
  }


  /**
   * getSimObject():  locate and return a simulation model of class 
   * 'classname' and version 'version' or higher.
   * <p>
   * The current preliminary implementation just ignores the requested
   * version number!
   * <p>
   * @exception SimObjectNotFoundException if the Object specified by
   *            classname and version number cannot be found
   * @exception SimObjectVersionNotFoundException if the version of
   *            SimObject cannot be found
   */
  public SimObject getSimObject( String classname, 
                                 long version, 
                                 Editor editor ) 
         throws SimObjectNotFoundException,
                SimObjectVersionNotFoundException {
    msg( "-W- DM.getSimObject(): ignoring requested version: " + version );
    return getSimObject( classname );
  }


  public SimObject getCopy( SimObject source, Editor editor ) {
    if (debug) msg( "-I- DesignManager.getCopy()..." );

    SimObject copy = source.copy();

    copy.setEditor( editor );
    //editorTable.put( copy, editor );
    // do bookkeeping here
    return copy;
  }


  public Design  getNewDesign( Editor editor ) {
    Design tmp = new Design();
    tmp.setEditor( editor );
    tmp.setVisible( true );
    //editorTable.put( tmp, editor );
    return tmp;
  }


  /**
   * getDesign: try to load a Design from file/URL "resourcename".
   * Set "toplevel" true for toplevel (visible) designs, and to false
   * for subdesigns.
   */
  public 
  Design getDesign( Editor editor, String resourcename, boolean toplevel ) 
  throws SimObjectNotFoundException
  {
    if (resourcename == null) {
      throw new SimObjectNotFoundException( "-E- Empty Design resourcename!" );
    }

    Design tmp = new Design();
    try {
      tmp.setEditor( editor );
      tmp.setVisible( toplevel );
      tmp.initialize( NameMangler.encodeWithUnicodeEscapes( resourcename ));
      //editorTable.put( tmp, editor );
    }
    catch( Exception e ) {
      msg( "-E- DesignManager.getDesign(): " + e );
      throw new SimObjectNotFoundException();
    }
    return tmp;
  }



  /**
   * this method checks whether the file/URL specified by "pathname"
   * already exists. If so, it renames the original file to a backup name.
   * <p>
   * Note that the current implementation only works for local files,
   * not for remote files or URLs, and not for files in JAR/Zip archives.
   * <p>
   * At the moment, the backup name is constructed by appending an
   * underscore character and an integer. For example, the backups for
   * Hades design "/examples/cla/adder8.hds" will be called
   * "/examples/cla/adder8.hds_1", "/examples/cla/ader8.hds_2", and so on.
   * We use the trivial algorithm for this, which becomes very slow
   * when the number of existing backup files is large.
   * <p>
   * @return  a flag indicating whether we created a backup
   */
  public boolean checkExistsMakeBackup( String pathname ) {
    File   orig    = null;
    File   tmpfile = null;
    String tmpname = null;
    try {
      orig    = new File( pathname );
      tmpfile = new File( pathname );

      if (!orig.exists()) return false;

      int i = 0;
      while( tmpfile.exists()) {
        tmpname = pathname + "_" + i;
        tmpfile = new File( tmpname );
        i++;
      }
      // here, we finally have a new filename
      //
      orig.renameTo( tmpfile ); 
      if (debug) msg( "-I- DM.checkExistsMakeBackup: renaming to " + tmpname );
    }
    catch( Exception e ) {
      msg( "-E- internal error in DesignManager.checkExistsMakeBackup: " );
      msg( "    pathname: `" + pathname + "' tmpname: `" + tmpname + "'" );
      msg( e.toString() );
      e.printStackTrace();
    }
    return true;
  }






  /**
   * try to find and open an InputStream for the given resource name.
   */
  public InputStream getResourceAsStream( String resourceName ) {
    if (debug) traceResourceAccess( resourceName );

    InputStream IS = null;

    //
    // phase 0: check whether we are running as an applet, if so,
    //          try to load resources via standard URLs: baseURL + resourceName
    //
    if (appletServerName != null) { 
      String fullName = null;
      try {

        msg( "-I- DesignManager.AppletWorkaround: loading " 
                 + appletServerName + " " 
                 + appletBaseDirName + " " 
                 + resourceName + " ..." );

        fullName = appletServerName + appletBaseDirName + resourceName;
        URL  url = new java.net.URL( fullName );

        IS      = url.openStream();
        if (IS != null) return IS;
      }
      catch( Exception e ) {
        jfig.utils.ExceptionTracer.trace( e );
      }
      System.err.println( 
             "-E- DesignManager.applet workaround failed for '"
             + fullName + "'" );
    }

    //
    // phase 1: try Java 1.1 (CLASSPATH) resource
    //
    try {
      IS = getClass().getResourceAsStream( resourceName );
      if (IS != null) return IS;
    }
    catch( Exception e ) {
      jfig.utils.ExceptionTracer.trace( e );
    }
    msg( "-W- DesignManager: could not find resource '"
           + resourceName + "', checking for a file now..." );

    //
    // phase 2: try to find a file
    //
    try {
      if (new File(resourceName).exists()) {
        IS = new FileInputStream( resourceName );
      }
      if (IS != null) return IS;
    }
    catch( Exception e ) {
      jfig.utils.ExceptionTracer.trace( e );
    }
    msg( "-W- DesignManager: could not find file '"
           + resourceName + "', checking for an URL now..." );


    //
    // phase 3: try a standard URL now...
    //
    try {
      IS = new java.net.URL( resourceName ).openStream();
      if (IS != null) return IS;
    }
    catch( Exception e ) {
      jfig.utils.ExceptionTracer.trace( e );
    }
    msg( "-E- DesignManager: could not find URL '"
           + resourceName + "'" );
    msg( "-E- Please check the resource/file/URL name"
           + " and your CLASSPATH settings!" );

    return null;
  }



  
  void traceResourceAccess( String resourceName ) {
    System.err.println( "-#- DesignManager.getResourceAsStream: " 
                      + resourceName );
  }


  /**************************************************************/ 
  /*   applet stuff and Netscape/MSIE bug workarouds ...        */
  /**************************************************************/ 

 

  private static String  appletServerName  = null;
  private static String  appletBaseDirName = null;


    
  public static String getAppletServerName() { return appletServerName; }

  public static void   setAppletServerName( String servername ) {
    System.out.println( "-#- DM.setAppletServerName: " + servername );
    appletServerName = servername;
  }

  public static String getAppletBaseDir() { return appletBaseDirName; }

  public static void   setAppletBaseDir( String name ) {
    appletBaseDirName = name;
  }

 
 
  /** 
   * create the Netscape/MSIE style dialog to select a file/URL name.
   * The dialog offers a textfield and the usual cancel/ok/browse buttons.
   * If necessary, we create a dummy (invisible) parent frame for the dialog.
   */
  private synchronized void
  checkCreateSelectFileOrURLDialog() {
    if (selectDialog == null) {
      checkCreateDummyFrame();
      selectDialog = new SelectFileOrURLDialog( frame );
    } 
  }

  /**
   * create a traditional file dialog for direct selection of filenames.
   * Most users seem to prefer this (understandably) over the 
   * more complex SelectFileOrURLDialog.
   * <p>
   * To avoid the f**cking JDK1.3.1/Windows crashes, we have to create
   * a new FileDialog each time. 
   * If necessary, we create a dummy (invisible) parent frame for the dialog.
   */
  private synchronized void
  checkCreateFileDialog() {
    if (fileDialog == null) {
      checkCreateDummyFrame();
      //  fileDialog = new FileDialog( frame ); // does not work on Win/JDK131
    }
    fileDialog = new FileDialog( frame, "AWT is broken"  );     // does work on Win/JDK131 :-(
  }


  /** 
   * check whether our dummyFrame instance variable is still null.
   * If so, create a new (dummy) Frame as a parent for the FileDialog
   * or SelectFileOrURLDialog windows.
   */
  private synchronized void
  checkCreateDummyFrame() {
    if (frame == null) {
      frame = new Frame( "dummy" );
      frame.setSize( 100, 100 );
      frame.show();
    }
  }

  /**
   * set a Frame which is used as a parent window for the FileDialog 
   * and SelectFileOrURLDialog windows. 
   */
  public void setDialogWindowsParentFrame( Frame frame ) {
    this.frame = frame;
  }


  /**
   * selectFileOrURLName: prompt the user for a file or URL name.
   * This method will return the null value if the user cancels the selection.
   * Other typical return values might be "adder16.hds", 
   * "http://www.some.server.org/path/data.rom" or
   * "file://home/joe.user/foo/bar.jpeg".
   * Note that this method will replace the platform-dependent 
   * separator char (e.g. the backslash '\' used by Windows)
   * in a selected filename with normal '/' slashes.
   * <p>
   * If non-null, dialogTitle is used for dialog window title,
   * defaultName is used as the initial file/URL name, and the
   * extension is used to create a matching hades.utils.EndsWithFilter.
   * The mode argument should be either java.awt.FileDialog.LOAD or 
   * java.awt.FileDialog.SAVE.
   * <p>
   * The current implementation uses AWT-based dialog windows (FileDialog
   * and hades.gui.SelectFileOrURLDialog), but this will use the
   * BrowserFrame JTree-based selection soon.
   * <p>
   * Question:  Why does the JDK deadlock or crashes on Windows, 
   * when this method is made synchronized?
   */
  //public synchronized
  public
  String selectFileOrURLName( String dialogTitle, 
                              String defaultName,
                              String extension,
                              int mode ) 
  {
    String name = null;

    // FIXME: we should also use SelectDialog instead of FileDialog 
    // when a) running as an applet, b) the defaultName startsWith http://
    // or another protocol, when c) defaultName startsWith resource:
    //
    if (SetupManager.getBoolean( 
             "Hades.DesignManager.UseSelectDialog", true )) {
       checkCreateSelectFileOrURLDialog(); 
       if (dialogTitle != null) {
         selectDialog.setTitle( dialogTitle );
       }
       if (extension != null) {
         selectDialog.setFilenameFilter( new EndsWithFilter( extension ));
       }
       if (defaultName != null) {
         selectDialog.setFilename( defaultName );
       }
       selectDialog.setMode( mode );
       selectDialog.show();

       // mode dialog here...
       name = selectDialog.getFilename();
    }
    else { // use traditional FileDialog directly, as most users prefer this

       checkCreateFileDialog(); 
       if (dialogTitle != null) {
         fileDialog.setTitle( dialogTitle );
       }
       if (extension != null) {
         fileDialog.setFilenameFilter( new EndsWithFilter( extension ));
       }
       if (defaultName != null) {
         setFileDialogDirectoryAndFilename( fileDialog, defaultName );
       }

       fileDialog.setMode( mode );
       fileDialog.show();

       // modal dialog here... wait a little bit after closing the dialog...
       jfig.utils.AntiDeadlock.sleep( 100 ); 

       if (fileDialog.getFile() != null) {
          name = fileDialog.getDirectory() + fileDialog.getFile();
          _lastFileDialogDirectory = fileDialog.getDirectory();
       }
    }

    // remove platform dependencies, always use '/' internally
    //
    if (name != null) {
      name = name.replace( File.separatorChar, '/' );
    }

    dbg( "-#- DM.selectFileOrURLName: " + name );
    return name;
  }


  /**
   * try to split resourceName (which always uses '/' as a separator char)
   * into directory and filename parts. 
   * Check that the resulting directory name denotes an existing directory.
   * If so, set the directory and filename in the fileDialog.
   */
  public void
  setFileDialogDirectoryAndFilename( FileDialog fileDialog, String resourceName)
  {

    String dirName  = null;
    String fileName = null;
dbg( "-#- sFDDAF: '" + resourceName + "'" );

    if (fileDialog == null) return;
    if (_lastFileDialogDirectory != null) {
       fileDialog.setDirectory( _lastFileDialogDirectory );
    }
dbg( "*** " + fileDialog.getDirectory() );

    // phase 1: resourceName always uses '/' as a separator char.
    // Therefore, we can just use the last '/' charactor to split the
    // resourceName into directory/base path and filename
    try {
      int lastSlash = resourceName.lastIndexOf( '/' );
      if (lastSlash >= 0) {
         dirName  = resourceName.substring( 0, lastSlash );
         fileName = resourceName.substring( lastSlash+1 );
      }
      else { // flat name, no directory
         fileName = resourceName;
      }
    }
    catch( Exception e ) {
      msg( "-#- Internal error in sFDDAF: " + e );
      e.printStackTrace();
    }

    // phase 2: now check that directory exists...
    //  
    if (dirName != null) {
      dirName = dirName.replace( '/', File.separatorChar );
      File dir = null;
      dir = new File( dirName );
      if ((dir != null) && (dir.exists())) {
           fileDialog.setDirectory( dirName );
      }
    }

    // phase 3: set fileName
    //
    if (fileName != null) fileDialog.setFile( fileName );

dbg( "-#- dirName  = '" + dirName + "'" );
dbg( "-#- fileName = '" + fileName + "'" );
  }

  /** 
   * utility method to "repair" a missing filename extension. 
   * 
   * Use this method to add the filename extension expected by Hades,
   * when the platform file dialog ignores FileNameFilter and the
   * user did not explicitly specify a filename extension.
   * <p> 
   * This method will first check the current value of the SetupManager
   * "Hades.DesignManager.AutoAddExtension" property. If not true,
   * the method returns the original "name" argument unmodified.
   * When enabled, this method will then check whether the string "name" 
   * already has an extension. If yes, it also returns the original string.
   * If not, it will add "extension" to "name".
   * <p>
   * For example, when called with "c:\tmp\adder" and ".hds" arguments,
   * this method will return "c:\tmp\adder.hds".
   * <p>
   * This is a dumb method, the check for an extension is made simply
   * by searching for a dot "." in the original "name". This algorithm is
   * not optimal when the user selects a name with multiple dots in it.
   */
  public String checkAndAddFilenameExtension( String name, String extension ) {
    if (!SetupManager.getBoolean("Hades.DesignManager.AutoAddExtension",true)){
      return name;
    }

    boolean nameHasDot = (name.indexOf( '.' ) >= 0);
    if (nameHasDot) return name;
    if (extension == null) return name;

    boolean extensionHasDot = (extension.indexOf( '.' ) >= 0);
    if (extensionHasDot) return name + extension;
    else                 return name + "." + extension;
  }


  /**
   * getInputStream: this method is the heart of DesignManager.
   * It tries to find the specified resource as error-tolerant as possible
   * and returns an InputStream for further processing.
   * It searches a variety of sources: local files, URLs, CLASSPATH
   * resources (zip://, file://, http://, resource:, local files).
   * Some examples for the supported pathnames:
     <pre>
     file://c:/users/foo/examples/bar.txt
     zip://c:/temp/foo.zip?/foo/foo2/foobar.hds
     hades/models/gates/And2.gif
     /hades/models/gates/And2.gif
     resource://hades/models/io/Ipin.sym
     ./foobar.rom
     </pre>
   * <p>
   * When searching for a local file, this method will automagically
   * replace any slash chars ('/') in pathname with the platform dependent
   * separator char.
   * <p>
   * When the first argument is non-null, it will also be used to
   * search for the given resource. For example, when the first
   * argument is a Hades Design, that Design resource name will be
   * used to search for pathname, too.
   * <p>
   */
  public InputStream getInputStream( Object parent, String pathname ) {
    InputStream  raw = null;
    String   tmpname = null;
    char   separator = File.separatorChar;

    if (debug) msg( "-#- DM.getIS: raw name '" + pathname + "'" );

    // phase 1: check for explicit protocol...
    //
    try {
      if (pathname.startsWith( "file://" )) {
        tmpname = pathname.substring(7).replace( '/', separator );
        raw = new FileInputStream( tmpname );
      }
      else if (pathname.startsWith( "file:" )) { // file, too few slashes...
        tmpname = pathname.substring(5).replace( '/', separator );
        raw = new FileInputStream( tmpname );
      }
      else if (pathname.startsWith( "zip://" )) { // Hades specific ZIP/JAR
        raw = getZipEntryInputStream( pathname );
      }
      else if (pathname.startsWith( "resource://" )) { // Classpath resource
        tmpname = pathname.substring( 11 );
        raw = getClass().getResourceAsStream( tmpname );
      }
      else if (pathname.indexOf( "://" ) > 0) { // looks like an URL
        URL rawURL = new URL( tmpname );
        raw = rawURL.openStream();
      }
    }
    catch( Exception e ) {
      msg( "-E- DM.getInputStream: while opening '" + pathname + "': " + e );
      msg( "-I- phase 1, internal name was: '" + tmpname + "'" );
      if (debug) e.printStackTrace();
    }
    if (raw!=null&&debug) msg( "-I- DM.getIS: tried URL: '" + tmpname + "'" );
    if (raw != null) return new BufferedInputStream( raw );


    //
    // phase 2: no protocol. Try to access a CLASSPATH resource,
    // adding the leading '/' if not already there
    //
    try {
      if (pathname.startsWith( "/" )) tmpname = pathname;
      else                            tmpname = "/" + pathname;
      raw = getClass().getResourceAsStream( tmpname );
    } 
    catch( Exception e ) {
      msg( "-E- DM.getInputStream: while opening '" + pathname + "': " + e );
      msg( "-I- phase 2, internal name was: '" + tmpname + "'" );
      if (debug) e.printStackTrace();
    }
    if (debug) msg( "-#- DM.getIS: tried resource: '" + tmpname + "'" );
    if (raw != null) return new BufferedInputStream( raw );



    // 
    // phase 2b: check for Netscape applet workaround, 
    // try to read single URL from the applet server
    // 
    // 
    if (appletServerName != null) {
      String fullName = null;
      InputStream IS  = null;
      try {
        msg( "-I- DesignManager.AppletWorkaround: trying '"
                 + appletServerName + " "
                 + appletBaseDirName + " "
                 + pathname + "' ..." );
        fullName = appletServerName + appletBaseDirName + pathname;
        URL  url = new java.net.URL( fullName );
        IS       = url.openStream();
        if (IS != null) return IS;
      }
      catch( Exception e ) {
        e.printStackTrace();
      }
      System.err.println(
        "-E- DesignManager.applet workaround failed for '" + fullName + "'" 
      );
    }


    //
    // phase 3: no protocol, no resource. Try to access a File
    //
    try {
      tmpname = pathname.replace( '/', separator );
      File file = new File( tmpname );
      if (file.exists() && file.canRead()) {
        raw = new FileInputStream( file );
      }

    } 
    catch( Exception e ) {
      msg( "-E- DM.getInputStream: while opening '" + pathname + "': " + e );
      msg( "-I- phase 3, internal name was: '" + tmpname + "'" );
      if (debug) e.printStackTrace();
    }
    if (debug) msg( "-I- DM.getIS: tried File: '" + tmpname + "'" );
    if (raw != null) return new BufferedInputStream( raw );




    //
    // phase 4: this is getting difficult now. Check if parent is a 
    // known object that can help us...
    //
    msg( "-W- DM.getInputStream: phase 4: '" + pathname + "'" );
     
    try {
      raw = lookForInputStream( parent, pathname );
    }
    catch( Exception e ) { 
      msg( "-E- DM.getInputStream: not found: '" + pathname + "'" );
      msg( "-E- exception is: " + e );
      // e.printStackTrace();
      return null;
    }
    return raw;
  }



  /**
   * try to get an InputStream for a resource called "pathname" for "object".
   * For example, object might be a Hades memory object, and "pathname"
   * specifies the data for the initial memory contents.
   * Or, object might be a hades.models.Design and "pathname"
   * names the Hades design file for the subdesign.
   */
  private InputStream lookForInputStream( Object object, String pathname ) 
    throws Exception
  {
    if (object == null) return null; // FIXME: this hack avoids recursion

    Design parent = null;
    String dirName = null;


    // some initial messages (for first debugging)
    //
    dbg( "-#- lookForInputStream: " );
    dbg( "    pathname='" + pathname + "'" );

    if (object instanceof Design) {
      //
      // this is only called, when initializing a new subdesign from
      // Design.initialize(). For consistency with the rest of the API,
      // the method is called with the this-parameter, which is to the
      // still un-initialized subdesign. 
      // Therefore, we check whether we have a valid (non-null) parent
      // Design, and use this to resolve the pathname.
      //
      parent = (Design) ((Design) object).getParent();
      dirName = parent.getResourceName();

      dbg( "    object is a Design '" + dirName + "'");

    }
    else if (object instanceof SimObject) {
      // 
      // this code path is used when trying to load external resources
      // for some SimObject, e.g. memory data.
      //
      dbg( "    object is a SimObject, " + object );

      // get the (sub-) Design the SimObject belongs to
      //
      parent = (Design) ((SimObject) object).getParent();
      dirName = parent.getResourceName();
      dbg( "    in parent Design '" + dirName + "'" );
    }
    else {
      // what kind of object is this?
      //
      dbg( "    object is " + object );
      dbg( "-E- I don't know how to handle this case." );
      return null;
    }

    // FIXME:  classpath-resource-hack.
    // Several of the Hades example designs use Java classpath resource names
    // like /hades/examples/cla/cla.hds to reference subdesigns or simobject
    // resources. To get those working with an invalid or null CLASSPATH,
    // we check whether the parent resourcename has a /hades in it,
    // and construct a corresonding name...
    // To handle cases with multiple levels of nested subdesigns, each of
    // which uses a resource name, we recurse up the hierarchy, until we
    // find a (sub-) design, whose name does not start with /hades...
    //
    if (pathname.startsWith( "/hades" )) {
      if (dirName.startsWith( "/hades" )) { // uuh, ooh, need nested hack
        return lookForInputStream( parent, pathname );
      }
      else {
        int index = dirName.lastIndexOf( "/hades" );
        if (index < 0) {
          dbg( "-E- internal: classpath hack, but no /hades in dirName" );
          return null;
        }
        String tmp = dirName.substring( 0, index ) + pathname;
        dbg( "-#- classpath hack: tmp '" + tmp + "'" );
        return getInputStream( null, tmp );
      }
    }


    // not a classpath-resource.
    // Try to load a relative pathname, which might also include ".."
    // to reference some other directory in the model/design tree.

    dbg( "-E- look for relative resource: NOT YET" );
    return getInputStreamFromRelativeName( dirName, pathname );
  }



  private InputStream 
  getInputStreamFromRelativeName( String parentname, String pathname ) {
    msg( "-#- DM.gSFRN: `" + parentname + "' `" + pathname + "'" );
    try {
      String parentDir = getDirectory( parentname );
      String fullname = parentDir + File.separator + pathname;
if (debug) msg( "-*- DM.gSFRN: " + fullname );
      return getInputStream( null, fullname );
    }
    catch( Exception e ) {
      msg( "-E- : " + e );
      return null;
    }
  }




  /**
   * take a file/URL/resource pathname and split it into tokens, one
   * substring per directory and one for the file/URL name including extension.
   */
  private static String[] tokenize( String pathname ) {
   // divide string into directory tokens
    //
    StringTokenizer ST = new StringTokenizer( pathname, File.separator );
    Vector tokenV = new Vector();
    while( ST.hasMoreTokens() ) {
      tokenV.addElement( ST.nextToken() );
    }
   
    String[] tokens = new String[ tokenV.size() ];
    for( int i=0; i < tokens.length; i++ ) {
      tokens[i] = (String) tokenV.elementAt(i);
    }

    if (debug) dumpTokens( tokens);
    return tokens;
  }


  public static void dumpTokens( String[] tokens ) {
    for( int i=0; i < tokens.length; i++ ) {
      System.out.print( " " +  tokens[i] );
    }
    System.out.println();
  }



  public OutputStream getOutputStream( Object o, String pathname ) {

    OutputStream OS = null;
    try {
      pathname = pathname.replace( '/', java.io.File.separatorChar );
      dbg( "-#- DM.getOutputStream: '" + pathname + "'" );
      return new FileOutputStream( pathname ); // "/tmp/hugo.txt" );
    }
    catch( Exception e ) {
      return System.out;
    }

  }



  /**
   * load, create and return an Image object from file/URL pathname.
   * As we expect the Image to be small and to be re-used, 
   * we cache the Images in an internal Hashtable, to speed up future
   * accesses to the same Image.
   * <p>
   * The raw image data is loaded via DesignManager.getInputStream()
   * into a ByteArrayOutputStream, and then converted into an Image
   * via java.awt.Toolkit.createImage().
   * Note that this method expects to be called with a valid image pathname.
   * No attempt is made to find an image by changing the pathname argument
   * (like changing name.hds to name.gif).
   * <p>
   * If the requested Image cannot be found, we try to return the
   * "/hades/manager/icons/null.gif" icon instead.
   */
  public java.awt.Image getThumbnail( String pathname ) {
    if (debug) msg( "-#- DM.getThumbnail: " + pathname );

    // look for cached thumbnail for 'pathname'
    //
    String  gifName = pathname;
    Image thumbnail = (Image) thumbnailTable.get( gifName );
    if (thumbnail != null) return thumbnail;

    // cache-miss: load the thumbnail raw data into a buffer, then
    // convert into an AWT Image
    //
    try {
      ByteArrayOutputStream baos = readStreamIntoBuffer( gifName );
      thumbnail = Toolkit.getDefaultToolkit().createImage(baos.toByteArray());
    }
    catch( Exception e ) {
      msg( "-E- DM.getThumbnail: not found from " + gifName );
      thumbnail = getDefaultThumbnail();
    }
    if (thumbnail != null) thumbnailTable.put( gifName, thumbnail );

    return thumbnail;
  }


  /**
   * construct a small Image to be used as a "not found" indicator
   * from "/hades/manager/icons/null.gif"
   */
  private java.awt.Image getDefaultThumbnail() {
    String _defaultName = "/hades/manager/icons/null.gif";
    try {
      ByteArrayOutputStream baos = readStreamIntoBuffer( _defaultName );
      return Toolkit.getDefaultToolkit().createImage(baos.toByteArray());
    }
    catch( Exception e ) {
      msg( "-F- DM.getDefaultThumbnail: not found from " + _defaultName );
      return null;
    }
  }







  /**
   * an utility method to read the specified resource into a (multiline)
   * String for further processing. 
   * This method returns null when the resource 'pathname' cannot be found.
   */
  public String readStreamIntoString( String pathname ) {
    ByteArrayOutputStream baos = readStreamIntoBuffer( pathname );
    if (baos != null) return baos.toString();
    else              return null;
  }


  /**
   * find the specified resource and read its contents into a
   * ByteArrayOutputStream buffer.
   * This method returns null when the resource 'pathname' cannot be found.
   * Internally, this method uses getInputStream( null, pathname )
   * to locate the resource.
   */
  public ByteArrayOutputStream readStreamIntoBuffer( String pathname ) {
    ByteArrayOutputStream baos = null;
    try {
      baos = new ByteArrayOutputStream();
      InputStream input = getInputStream( null, pathname );

      if (input != null) {
        byte buffer[] = new byte[4096];
        int length =0, count = 0;
        while( (length = input.read( buffer, 0, buffer.length )) >= 0) {
          // System.out.println( "writing " + length + " bytes..." );
          baos.write( buffer, 0, length );
          count += length;
        }
      }
      baos.close();
      return baos;
    }
    catch( Throwable t ) {
      t.printStackTrace();
    }
    return null;
  }




  /**************************************************************/ 
  /*   ZIP file stuff                                           */
  /**************************************************************/ 



  public void registerZipFile( ZipFile zipfile ) {
    _zipfileTable.put( zipfile.getName(), zipfile );

    msg( "-I- registerZipFile: " + zipfile.getName() );
  }



  /**
   * given a pathname of the form "zip://hugo.zip?/path/to/file.hds"
   * locate the corresponding zipfile (which should have been registered
   * by calling registerZipFile), and return an InputStream for the
   * specified zip entry.
   */
  public InputStream getZipEntryInputStream( String pathname ) {
    try {
      int       index = pathname.indexOf( '?' );
      String  zipname = pathname.substring( 6, index );
      String filename = pathname.substring( index+2 ); // strip "?/"

      if (debug) 
        msg( "-#- DM.getZEIS: zip='" + zipname + "' file='" + filename +"'" );

      ZipFile   zipfile = (ZipFile) _zipfileTable.get( zipname );
      InputStream input = zipfile.getInputStream( zipfile.getEntry( filename ));
      return input;
    }
    catch( Exception e ) {
      msg( "-E- getZipEntryInputStream failed for pathname: " + pathname );
      return null;
    }
  }



  /**************************************************************/ 
  /*   utitity methods                                          */
  /**************************************************************/ 


  public static
  String getFilename( String pathname ) {
    //File f = new File( pathname ).getAbsoluteFile();
    //return f.getName();
    //since getAbsoluteFile requires JDK1.2, try the following
    return new File( pathname ).getAbsolutePath();
  }


  public static 
  String getDirectory( String pathname ) {
    //File f = new File( pathname ).getAbsoluteFile();
    //return f.getParent();
    //since getAbsoluteFile requires JDK1.2, try the following
    String tmp = pathname.replace( '/', File.separatorChar );
    String parentname = new File( tmp ).getParent();
    return new File( parentname ).getAbsolutePath();
  }


  /**
   * get the extension (starting with the last "." dot char) 
   * of this pathname, or null.
   * For example, called with "/tmp/hades/adder.hds", this method
   * will return ".hds".
   */
  public static 
  String getExtension( String pathname ) {
    int index = pathname.lastIndexOf( "." );
    if (index < 0) return null;
    else return pathname.substring(index);
  }


  /**
   * change the extension of original filename "source" to the
   * "newExtension". If the original name has no extension at all,
   * add the new extension. The extension is interpreted as the
   * part of a name beginning with the last "." dot char.
   * <p>
   * For example, calling this method with arguments
   * "/tmp/design/traffic.hds" and ".sym"
   * will create the name "/tmp/design/traffic.sym".
   * This dumb method is not designed to handle illegal or null arguments.
   */
  public static
  String changeExtension( String source, String newExtension ) {
    int index = source.lastIndexOf( "." );
    if (index < 0) return (source + newExtension);
    else           return (source.substring(0,index) + newExtension);
  }


  


  public void setDebug( boolean b ) { debug = b; }
  public boolean getDebug() { return debug; }

  public static void dbg( String msg ) {
    if (SetupManager.getBoolean("Hades.DesignManager.VerboseMessages",false)) {
      System.out.println( msg );
    }
  }

  public void msg( String msg ) { 
    System.out.println( msg );
    //hades.gui.Console.getConsole().message( msg );
  }


  public String toString() {
    return "HADES DesignManager 0.1";
  }
}

