/* MakePieChart.java - jfig.utils.MakePieChart 
 *
 * demonstrate how to expand jfig with custom components - this one
 * creates pie charts 
 *
 * 22.08.00 - use the new attribute setters
 * 16.08.00 - working, and lots of comments
 * 27.02.00 - first version
 *
 */

package jfig.utils;                                // use your own package here


import java.awt.*;

import jfig.gui.*;
import jfig.objects.*;
import jfig.canvas.*;



public class MakePieChart {

  static final double _2PI = 2 * Math.PI;

  ModularEditor  editor = null; // the jFIG editor we're using


  /**
   * construct a slice of an pie chart with the following parameters:
   * radius - the pie radius in jfig units (2400dpi)
   * ratio  - the size of this slice, as a fraction of the total 0 .. 1.0
   * angle  - the start angle of this slice
   * Note that you can always set the color, attributes, angle etc. 
   * afterwards by calling standard jfig functions on the FigXXX objects.
   * This method creates FigPolylines to approximate the pie segments
   * (instead of FigArc) to demonstrate setPoints() etc.
   */
  public FigObject makePieSlice( Point center,
                                 double radius, 
                                 double ratio, 
                                 double angle,
                                 Color  color )
  {
    // we approximate pie segments by 200 points per full circle
    //
    int    n_points = 2 + (int) (200.0 * ratio); 
    double deltaphi = ratio * _2PI / (n_points-2);
    Point  points[] = new Point[n_points];

    points[0] = new Point( center.x, center.y );

    double phi = angle;
    for( int i=1; i < n_points; i++) {
      points[i] = new Point( center.x + (int) (radius*Math.cos(phi)+0.5),
                             center.y + (int) (radius*Math.sin(phi)+0.5) 
                           );
      phi += deltaphi;
    }

    FigPolyline poly = new FigPolyline();
    poly.setIsClosed( true );
    poly.setLineColor( Color.white );
    poly.setLineWidth( 1 );

    poly.setFillStyle( FigAttribs.SOLID_FILL );
    poly.setFillColor( color );

    poly.setTrafo( editor.getObjectCanvas().getTrafo() );
    poly.setPoints( points );
    return poly;
  }



  public FigCompound makePieChart( double ratios[] )
  {
    // many jfig method need references to the editor coordinate transformation
    // or the currently active object attributes (line width, fill color, etc.)
    //
    FigCompound  compound = new FigCompound();
    compound.setTrafo( editor.getObjectCanvas().getTrafo() );

    Point        center   = new Point( 4800, 4800 );
    double       radius   = 2400;         // 2 inch default
    double       angle    = 0.0;
    
    for( int i=0; i < ratios.length; i++ ) {
      Color     color = ColorCache.getColorCache().get( i%10 );
      FigObject slice = makePieSlice( center, radius, ratios[i], angle, color );
      compound.addMember( slice );

      angle += ratios[i] * _2PI;
    }
    return compound;
  }


  

  public static void main( String argv[] ) {
    MakePieChart MPC   = new MakePieChart();

    // create a jfig editor, this automatically loads the global and user
    // property settings, if any. And we are not interested in verbose error
    // messages.
    //
    jfig.utils.ExceptionTracer.setEnabled( false );
    MPC.editor = new ModularEditor();


    // create two pie charts from the raw data
    //
    double[] sales     = { 0.1, 0.2, 0.3, 0.1, 0.15, 0.15 };
    double[] market    = { 0.0, 0.55, 0.23, 0.22 };
    FigCompound chart1 = MPC.makePieChart( sales );
    FigCompound chart2 = MPC.makePieChart( market );

    // move the charts, 2400 dpi: 2400 = 1 inch, 7200 = 3 inch, etc.
    //
    chart1.move( 7200, 2400 );         
    chart2.move(    0, 2400 ); 

    // add the chart objects to the editors object list
    //
    MPC.editor.insertIntoObjectList( chart1 );
    MPC.editor.insertIntoObjectList( chart2 );
    MPC.editor.doRedraw();
  }

}
